@@ -21,6 +21,8 @@
 #include <cstdlib>
 #include <config.h>
 #include <gtkmm/box.h>
+#include <gtkmm/messagedialog.h>
+#include <gtkmm/actiongroup.h>
 #include <gdk/gdkkeysyms.h>
 #include <sigc++/retype_return.h>
 #include "DCCList.h"
@@ -52,9 +54,10 @@ MainWindow::MainWindow(bool autoconnect)
           move(x, y);
     
     setupMenus();
+    initializeTagTable();
     Gtk::VBox *vbox = manage(new Gtk::VBox());
 
-    vbox->pack_start(_menubar, Gtk::PACK_SHRINK);
+    vbox->pack_start(*_uimanager->get_widget("/MenuBar"), Gtk::PACK_SHRINK);
     vbox->pack_start(_notebook, Gtk::PACK_EXPAND_WIDGET);
     vbox->pack_start(_statusbar, Gtk::PACK_SHRINK);
 
@@ -62,7 +65,7 @@ MainWindow::MainWindow(bool autoconnect)
     show_all();
 
     if (_app.options.hidemenu)
-          _menubar.hide();
+          _uimanager->get_widget("/MenuBar")->hide();
 
     if (_app.options.hidestatusbar)
           _statusbar.hide();
@@ -70,10 +73,12 @@ MainWindow::MainWindow(bool autoconnect)
     if (!_app.cfgservers.hasAutoConnects() || !autoconnect) {
         // Construct initial tab
         newServerTab();
+        openServerWindow();
     } else {
         // Auto-connect to servers.
         _app.autoConnect();
     }
+    _notebook.getCurrent()->getEntry().grab_focus();
 }
 
 MainWindow::~MainWindow()
@@ -97,37 +102,8 @@ MainWindow::~MainWindow()
     AppWin = 0;
 }
 
-void MainWindow::displayMessage(const ustring& msg, FE::Destination d, bool shouldHighlight)
-{
-
-    if (d == FE::CURRENT) {
-        Tab *tab = _notebook.getCurrent();
-
-        if (tab) {
-            tab->getText() << msg;
-
-            if (shouldHighlight)
-                  tab->highlightActivity();
-        }
-
-    } else if (d == FE::ALL) {
-        vector<Tab*> tabs;
-        vector<Tab*>::const_iterator i;
-        _notebook.Tabs(tabs);
-
-        for (i = tabs.begin(); i != tabs.end(); ++i) {
-            (*i)->getText() << msg;
-
-            if (shouldHighlight)
-                  (*i)->highlightActivity();
-        }
-    
-    }
-}
-
 void MainWindow::displayMessage(const ustring& msg, FE::Destination d, ServerConnection *conn, bool shouldHighlight)
 {
-
     if (d == FE::CURRENT) {
         Tab *tab = _notebook.getCurrent(conn);
 
@@ -141,7 +117,7 @@ void MainWindow::displayMessage(const ustring& msg, FE::Destination d, ServerCon
     } else if (d == FE::ALL) {
         vector<Tab*> tabs;
         vector<Tab*>::const_iterator i;
-        _notebook.findTabs(conn, tabs);
+        _notebook.findTabs(tabs, conn);
 
         for (i = tabs.begin(); i != tabs.end(); ++i) {
             (*i)->getText() << msg;
@@ -160,7 +136,7 @@ void MainWindow::displayMessage(const ustring& msg, ChannelBase& chan, ServerCon
     // created on join) - there is also a hack here to ensure that it's not
     // a channel
     char p = chan.getName().at(0);
-    if (!tab && (p != '#' && p != '&' && p != '!' && p != '+'))
+    if (!tab && !conn->isChannelPrefix(p))
         tab = _notebook.addTab(Tab::QUERY, chan.getName(), conn);
 
     if (tab) {
@@ -270,7 +246,7 @@ void MainWindow::connected(ServerConnection* conn)
     vector<Tab*> tabs;
     vector<Tab*>::const_iterator i;
 
-    _notebook.findTabs(conn, tabs);
+    _notebook.findTabs(tabs, conn);
 
     for (i = tabs.begin(); i != tabs.end(); ++i)
           if ((*i)->isType(Tab::QUERY))
@@ -281,7 +257,7 @@ void MainWindow::disconnected(ServerConnection* conn)
 {
     vector<Tab*> tabs;
 
-    _notebook.findTabs(conn, tabs);
+    _notebook.findTabs(tabs, conn);
 
     std::for_each(tabs.begin(), tabs.end(), std::mem_fun(&Tab::setInActive));
 }
@@ -290,16 +266,20 @@ void MainWindow::newTab(ServerConnection *conn)
 {
     ustring name = _("server");
     conn->Session.servername = name;
-    Tab *tab = _notebook.addTab(Tab::SERVER, name, conn);
-    tab->setType(Tab::SERVER);
-
-    // XXX: this is a hack for a "bug" in the gtkmm code which makes the
-    // application crash in the start when no pages exists, even though we
-    // added one above... doing set_current_page(0) will somehow add it fully.
-    if (_notebook.get_current_page() == -1) {
-        _notebook.set_current_page(0);
+    Tab *tab = 0;
+    Tab *currenttab = getNotebook().getCurrent();
+    if (currenttab != 0 &&
+            !currenttab->getConn()->Session.isConnected &&
+            !currenttab->getConn()->Session.isConnecting &&
+            currenttab->isType(Tab::SERVER)) {
+        tab = currenttab;
+        tab->setName(name);
+        tab->setConn(conn);
+    } else {
+        tab = _notebook.addTab(Tab::SERVER, name, conn);
     }
-    tab->setInActive();
+
+    tab->setType(Tab::SERVER);
 }
 
 Tab* MainWindow::newServerTab()
@@ -344,7 +324,7 @@ void MainWindow::localeError(bool tried_custom_encoding)
     msg += _("\n\n(Note: You'll only see this warning once per LostIRC session)");
 
 
-    Gtk::MessageDialog mdialog(*this, msg, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK);
+    Gtk::MessageDialog mdialog(*this, msg, false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK);
     mdialog.run();
 }
 
@@ -373,89 +353,76 @@ void MainWindow::openDccWindow()
         _dccwin = dialog;
     }
 }
+
 void MainWindow::setupMenus()
 {
-    { // First menu.
-        Gtk::Menu::MenuList& menulist = _firstmenu.items();
+    _uimanager = Gtk::UIManager::create();
+    Glib::RefPtr<Gtk::ActionGroup> group = Gtk::ActionGroup::create();
 
-        menulist.push_back(Gtk::Menu_Helpers::MenuElem(
-                    _("New Server Tab"),
-                    Gtk::Menu::AccelKey("<control>n"),
-                    SigC::hide_return(SigC::slot(*this, &MainWindow::newServerTab))));
+    group->add(Gtk::Action::create("LostIRCMenu", _("_LostIRC")));
+    group->add(Gtk::Action::create("NewServerTab", _("_New Server Tab")),
+            Gtk::AccelKey("<control>n"),
+            sigc::hide_return(sigc::mem_fun(*this, &MainWindow::newServerTab)));
+    group->add(Gtk::Action::create("ClearWindow", _("Clear Window")),
+            sigc::mem_fun(_notebook, &MainNotebook::clearWindow));
+    group->add(Gtk::Action::create("ClearAllWindows", _("Clear All Windows")),
+            sigc::mem_fun(_notebook, &MainNotebook::clearAll));
+    group->add(Gtk::Action::create("CloseCurrentTab", _("Close Current Tab")),
+            Gtk::AccelKey("<control>w"),
+            sigc::mem_fun(*this, &MainWindow::closeCurrentTab));
+    group->add(Gtk::Action::create("Quit", Gtk::Stock::QUIT),
+            sigc::mem_fun(*this, &Gtk::Window::hide));
 
-        menulist.push_back(Gtk::Menu_Helpers::MenuElem(
-                    _("Clear Window"),
-                    SigC::slot(_notebook, &MainNotebook::clearWindow)));
+    group->add(Gtk::Action::create("ViewMenu", _("_View")));
+    group->add(Gtk::Action::create("MenuItem", _("_Menubar")), Gtk::AccelKey("<control>m"), sigc::mem_fun(*this, &MainWindow::hideMenu));
+    group->add(Gtk::Action::create("StatusBar", _("Status_bar")), Gtk::AccelKey("<control>b"), sigc::mem_fun(*this, &MainWindow::hideStatusbar));
+    group->add(Gtk::Action::create("UserList", _("_User List")), Gtk::AccelKey("<control>l"), sigc::mem_fun(*this, &MainWindow::hideNickList));
+    group->add(Gtk::Action::create("ServerList", _("_Server List")), Gtk::AccelKey("<control>s"), sigc::mem_fun(*this, &MainWindow::openServerWindow));
+    group->add(Gtk::Action::create("DCCTransfers", _("_DCC Transfers")), Gtk::AccelKey("<control>d"), sigc::mem_fun(*this, &MainWindow::openDccWindow));
+    group->add(Gtk::Action::create("Preferences", Gtk::Stock::PREFERENCES), Gtk::AccelKey("<control>p"), sigc::mem_fun(*this, &MainWindow::openPrefs));
 
-        menulist.push_back(Gtk::Menu_Helpers::MenuElem(
-                    _("Clear All Windows"),
-                    SigC::slot(_notebook, &MainNotebook::clearAll)));
+    group->add(Gtk::Action::create("HelpMenu", _("_Help")));
+    group->add(Gtk::Action::create("Introduction", _("_Introduction")), sigc::mem_fun(*this, &MainWindow::openHelpIntro));
+    group->add(Gtk::Action::create("About", _("_About")), sigc::mem_fun(*this, &MainWindow::openAboutWindow));
 
-        menulist.push_back(Gtk::Menu_Helpers::MenuElem(
-                    _("Close Current Tab"),
-                    Gtk::Menu::AccelKey("<control>w"),
-                    SigC::slot(*this, &MainWindow::closeCurrentTab)));
+    _uimanager->insert_action_group(group);
+    add_accel_group(_uimanager->get_accel_group());
 
+    try {
 
-        menulist.push_back(Gtk::Menu_Helpers::SeparatorElem::SeparatorElem());
-        menulist.push_back(
-                Gtk::Menu_Helpers::StockMenuElem(Gtk::Stock::QUIT,
-                    Gtk::Menu::AccelKey("<control>q"),
-                    SigC::slot(*this, &Gtk::Window::hide)));
+        Glib::ustring ui_info = 
+                "<ui>"
+                "  <menubar name='MenuBar'>"
+                "    <menu action='LostIRCMenu'>"
+                "      <menuitem action='NewServerTab'/>"
+                "      <menuitem action='ClearWindow'/>"
+                "      <menuitem action='ClearAllWindows'/>"
+                "      <menuitem action='CloseCurrentTab'/>"
+                "      <separator/>"
+                "      <menuitem action='Quit'/>"
+                "    </menu>"
+                "    <menu action='ViewMenu'>"
+                "      <menuitem action='MenuItem'/>"
+                "      <menuitem action='StatusBar'/>"
+                "      <separator/>"
+                "      <menuitem action='UserList'/>"
+                "      <menuitem action='ServerList'/>"
+                "      <menuitem action='DCCTransfers'/>"
+                "      <menuitem action='Preferences'/>"
+                "    </menu>"
+                "    <menu action='HelpMenu'>"
+                "      <menuitem action='Introduction'/>"
+                "      <separator/>"
+                "      <menuitem action='About'/>"
+                "    </menu>"
+                "  </menubar>"
+                "</ui>";
+
+        _uimanager->add_ui_from_string(ui_info);
+
+    } catch (const Glib::Error& ex) {
+        std::cerr << "Building menus failed: " << ex.what() << std::endl;
     }
-
-    { // View menu.
-        Gtk::Menu::MenuList& menulist = _viewmenu.items();
-
-
-        menulist.push_back(Gtk::Menu_Helpers::MenuElem(
-                    _("_Menubar"),
-                    Gtk::Menu::AccelKey("<control>m"),
-                    SigC::slot(*this, &MainWindow::hideMenu)));
-
-        menulist.push_back(Gtk::Menu_Helpers::MenuElem(
-                    _("Status_bar"),
-                    Gtk::Menu::AccelKey("<control>b"),
-                    SigC::slot(*this, &MainWindow::hideStatusbar)));
-
-        menulist.push_back(Gtk::Menu_Helpers::SeparatorElem::SeparatorElem());
-
-        menulist.push_back(Gtk::Menu_Helpers::CheckMenuElem(
-                    _("User _List"),
-                    Gtk::Menu::AccelKey("<control>l"),
-                    SigC::slot(*this, &MainWindow::hideNickList)));
-
-        menulist.push_back(Gtk::Menu_Helpers::MenuElem(
-                    _("_Server List"),
-                    Gtk::Menu::AccelKey("<control>s"),
-                    SigC::slot(*this, &MainWindow::openServerWindow)));
-        menulist.push_back(Gtk::Menu_Helpers::MenuElem(
-                    _("_DCC Transfers"),
-                    Gtk::Menu::AccelKey("<control>d"),
-                    SigC::slot(*this, &MainWindow::openDccWindow)));
-        menulist.push_back(
-                Gtk::Menu_Helpers::StockMenuElem(Gtk::Stock::PREFERENCES,
-                    Gtk::Menu::AccelKey("<control>p"),
-                    SigC::slot(*this, &MainWindow::openPrefs)));
-
-
-    }
-
-    { // Help menu.
-        Gtk::Menu::MenuList& menulist = _helpmenu.items();
-
-        menulist.push_back(Gtk::Menu_Helpers::MenuElem(
-                    _("_Introduction"), SigC::slot(*this, &MainWindow::openHelpIntro)));
-
-        menulist.push_back(Gtk::Menu_Helpers::SeparatorElem::SeparatorElem());
-
-        menulist.push_back(Gtk::Menu_Helpers::MenuElem(
-                    _("_About"), SigC::slot(*this, &MainWindow::openAboutWindow)));
-    }
-
-    _menubar.items().push_back(Gtk::Menu_Helpers::MenuElem(_("_LostIRC"), _firstmenu));
-    _menubar.items().push_back(Gtk::Menu_Helpers::MenuElem(_("_View"), _viewmenu));
-    _menubar.items().push_back(Gtk::Menu_Helpers::MenuElem(_("_Help"), _helpmenu));
 }
 
 
@@ -463,9 +430,9 @@ void MainWindow::hideMenu()
 {
     _app.options.hidemenu = !_app.options.hidemenu;
     if (_app.options.hidemenu)
-          _menubar.hide();
+          _uimanager->get_widget("/MenuBar")->hide();
     else
-          _menubar.show();
+          _uimanager->get_widget("/MenuBar")->show();
 }
 
 void MainWindow::hideStatusbar()
@@ -482,7 +449,7 @@ void MainWindow::hideNickList()
     _app.options.hidenicklist = !_app.options.hidenicklist;
     vector<Tab*> tabs;
 
-    _notebook.Tabs(tabs);
+    _notebook.findTabs(tabs);
     std::for_each(tabs.begin(), tabs.end(), std::mem_fun(&Tab::toggleNickList));
 }
 
@@ -504,9 +471,9 @@ void MainWindow::openHelpIntro()
     if (_helpwin.get()) {
           _helpwin->present();
     } else {
-        std::auto_ptr<Gtk::MessageDialog> dialog(new Gtk::MessageDialog(_("LostIRC Quick Introduction\n\nThis help window is a quick guide to get you going with LostIRC.\nMove this window away from the LostIRC window, and use it as a quick reference window until you know the general idea.\n\nYou can connect to a server using:\n    /SERVER <hostname / ip>\n\n...and then join a channel:\n    /JOIN <channel-name>\n\nA list of all commands are available with:\n    /COMMANDS\n\nAnd you should really check out the list of key bindings:\n    /KEYBINDINGS"), Gtk::MESSAGE_INFO, Gtk::BUTTONS_CLOSE, false));
+        std::auto_ptr<Gtk::MessageDialog> dialog(new Gtk::MessageDialog(_("LostIRC Quick Introduction\n\nThis help window is a quick guide to get you going with LostIRC.\nMove this window away from the LostIRC window, and use it as a quick reference window until you know the general idea.\n\nYou can connect to a server using:\n    /SERVER <hostname / ip>\n\n...and then join a channel:\n    /JOIN <channel-name>\n\nA list of all commands are available with:\n    /COMMANDS\n\nAnd you should really check out the list of key bindings:\n    /KEYBINDINGS"), false, Gtk::MESSAGE_INFO, Gtk::BUTTONS_CLOSE, false));
 
-        dialog->signal_response().connect(slot(*this, &MainWindow::hideHelpIntro));
+        dialog->signal_response().connect(sigc::mem_fun(*this, &MainWindow::hideHelpIntro));
         dialog->show();
 
         _helpwin = dialog;
@@ -518,9 +485,9 @@ void MainWindow::openAboutWindow()
     if (_aboutwin.get()) {
           _aboutwin->present();
     } else {
-        std::auto_ptr<Gtk::MessageDialog> dialog(new Gtk::MessageDialog(_("LostIRC "VERSION), Gtk::MESSAGE_INFO, Gtk::BUTTONS_OK, false));
+        std::auto_ptr<Gtk::MessageDialog> dialog(new Gtk::MessageDialog(_("LostIRC "VERSION), false, Gtk::MESSAGE_INFO, Gtk::BUTTONS_OK, false));
 
-        dialog->signal_response().connect(slot(*this, &MainWindow::hideAboutWindow));
+        dialog->signal_response().connect(sigc::mem_fun(*this, &MainWindow::hideAboutWindow));
         dialog->show();
 
         _aboutwin = dialog;
@@ -593,14 +560,103 @@ bool MainWindow::on_key_press_event(GdkEventKey* e)
               _notebook.prev_page();
         else if (e->keyval == GDK_Right)
               _notebook.next_page();
-    }
-    if (e->keyval == GDK_Page_Up) {
-        _notebook.getCurrent()->getText().scrollUpPage();
+    } else {
+        if (e->keyval == GDK_Page_Up) {
+            _notebook.getCurrent()->getText().scrollUpPage();
 
-    } else if (e->keyval == GDK_Page_Down) {
-        _notebook.getCurrent()->getText().scrollDownPage();
+        } else if (e->keyval == GDK_Page_Down) {
+            _notebook.getCurrent()->getText().scrollDownPage();
+        }
     }
 
     Gtk::Window::on_key_press_event(e);
     return false;
 }
+
+void MainWindow::addToTable(Glib::ustring name, Glib::RefPtr<Gtk::TextTagTable> table, const Glib::ustring& colorname)
+{
+    Glib::RefPtr<Gtk::TextTag> tag_fg = Gtk::TextTag::create(Glib::ustring("f")+name);
+    tag_fg->property_foreground() = colorname;
+    table->add(tag_fg);
+
+    Glib::RefPtr<Gtk::TextTag> tag_bg = Gtk::TextTag::create(Glib::ustring("b")+name);
+    tag_bg->property_background() = colorname;
+    table->add(tag_bg);
+}
+
+void MainWindow::initializeTagTable()
+{
+    _tag_table1 = Gtk::TextTagTable::create();
+    _tag_table2 = Gtk::TextTagTable::create();
+
+    addToTable("0", _tag_table1, App->colors1.color0);
+    addToTable("1", _tag_table1, App->colors1.color1);
+    addToTable("2", _tag_table1, App->colors1.color2);
+    addToTable("3", _tag_table1, App->colors1.color3);
+    addToTable("4", _tag_table1, App->colors1.color4);
+    addToTable("5", _tag_table1, App->colors1.color5);
+    addToTable("6", _tag_table1, App->colors1.color6);
+    addToTable("7", _tag_table1, App->colors1.color7);
+    addToTable("8", _tag_table1, App->colors1.color8);
+    addToTable("9", _tag_table1, App->colors1.color9);
+    addToTable("10", _tag_table1, App->colors1.color10);
+    addToTable("11", _tag_table1, App->colors1.color11);
+    addToTable("12", _tag_table1, App->colors1.color12);
+    addToTable("13", _tag_table1, App->colors1.color13);
+    addToTable("14", _tag_table1, App->colors1.color14);
+    addToTable("15", _tag_table1, App->colors1.color15);
+    addToTable("16", _tag_table1, App->colors1.color16);
+    addToTable("17", _tag_table1, App->colors1.color17);
+    addToTable("18", _tag_table1, App->colors1.color18);
+    addToTable("19", _tag_table1, App->colors1.color19);
+
+    addToTable("0", _tag_table2, App->colors2.color0);
+    addToTable("1", _tag_table2, App->colors2.color1);
+    addToTable("2", _tag_table2, App->colors2.color2);
+    addToTable("3", _tag_table2, App->colors2.color3);
+    addToTable("4", _tag_table2, App->colors2.color4);
+    addToTable("5", _tag_table2, App->colors2.color5);
+    addToTable("6", _tag_table2, App->colors2.color6);
+    addToTable("7", _tag_table2, App->colors2.color7);
+    addToTable("8", _tag_table2, App->colors2.color8);
+    addToTable("9", _tag_table2, App->colors2.color9);
+    addToTable("10", _tag_table2, App->colors2.color10);
+    addToTable("11", _tag_table2, App->colors2.color11);
+    addToTable("12", _tag_table2, App->colors2.color12);
+    addToTable("13", _tag_table2, App->colors2.color13);
+    addToTable("14", _tag_table2, App->colors2.color14);
+    addToTable("15", _tag_table2, App->colors2.color15);
+    addToTable("16", _tag_table2, App->colors2.color16);
+    addToTable("17", _tag_table2, App->colors2.color17);
+    addToTable("18", _tag_table2, App->colors2.color18);
+    addToTable("19", _tag_table2, App->colors2.color19);
+
+    // Create a underlined-tag.
+    Glib::RefPtr<Gtk::TextTag> underlinetag1 = Gtk::TextTag::create("U");
+    underlinetag1->property_underline() = Pango::UNDERLINE_SINGLE;
+    underlinetag1->property_foreground() = App->colors1.color0;
+    _tag_table1->add(underlinetag1);
+
+    Glib::RefPtr<Gtk::TextTag> underlinetag2 = Gtk::TextTag::create("U");
+    underlinetag2->property_underline() = Pango::UNDERLINE_SINGLE;
+    underlinetag2->property_foreground() = App->colors1.color0;
+    _tag_table2->add(underlinetag2);
+
+    // Create a bold-tag
+    Glib::RefPtr<Gtk::TextTag> boldtag1 = Gtk::TextTag::create("B");
+    boldtag1->property_weight() = Pango::WEIGHT_BOLD;
+    _tag_table1->add(boldtag1);
+
+    Glib::RefPtr<Gtk::TextTag> boldtag2 = Gtk::TextTag::create("B");
+    boldtag2->property_weight() = Pango::WEIGHT_BOLD;
+    _tag_table2->add(boldtag2);
+
+    if (_app.options.colorscheme == 1) {
+        background_color = _app.colors1.bgcolor;
+        _current_tag_table = _tag_table1;
+    } else {
+        background_color = _app.colors2.bgcolor;
+        _current_tag_table = _tag_table2;
+    }
+
+}
